@@ -1,11 +1,12 @@
 {- git-annex command
  -
- - Copyright 2010-2019 Joey Hess <id@joeyh.name>
+ - Copyright 2010-2022 Joey Hess <id@joeyh.name>
  -
  - Licensed under the GNU AGPL version 3 or higher.
  -}
 
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE OverloadedStrings #-}
 
 module Command.Fsck where
 
@@ -15,10 +16,11 @@ import qualified Remote
 import qualified Types.Backend
 import qualified Backend
 import Annex.Content
-import qualified Annex.Content.Direct as Direct
-import Annex.Direct
+import Annex.Content.Presence
+import Annex.Content.Presence.LowLevel
 import Annex.Perms
 import Annex.Link
+import Annex.Version
 import Logs.Location
 import Logs.Trust
 import Logs.Activity
@@ -28,21 +30,25 @@ import Annex.NumCopies
 import Annex.UUID
 import Annex.ReplaceFile
 import Utility.DataUnits
-import Config
 import Utility.HumanTime
 import Utility.CopyFile
 import Git.FilePath
 import Utility.PID
+import Utility.InodeCache
+import Annex.InodeSentinal
 import qualified Database.Keys
 import qualified Database.Fsck as FsckDb
 import Types.CleanupActions
 import Types.Key
 import Types.ActionItem
+import qualified Utility.RawFilePath as R
 
 import Data.Time.Clock.POSIX
 import System.Posix.Types (EpochTime)
 import qualified Data.Set as S
 import qualified Data.Map as M
+import Data.Either
+import qualified System.FilePath.ByteString as P
 
 cmd :: Command
 cmd = withGlobalOptions [jobsOption, jsonOptions, annexedMatchingOptions] $
@@ -82,7 +88,7 @@ optParser desc = FsckOptions
 			( long "more" <> short 'm'
 			<> help "continue an incremental fsck"
 			)
-		<|> (ScheduleIncrementalO <$> option (str >>= parseDuration)
+		<|> (ScheduleIncrementalO <$> option (eitherReader parseDuration)
 			( long "incremental-schedule" <> metavar paramTime
 			<> help "schedule incremental fscking"
 			))
@@ -93,36 +99,44 @@ seek o = startConcurrency commandStages $ do
 	u <- maybe getUUID (pure . Remote.uuid) from
 	checkDeadRepo u
 	i <- prepIncremental u (incrementalOpt o)
-	withKeyOptions (keyOptions o) False
+	let seeker = AnnexedFileSeeker
+		{ startAction = start from i
+		, checkContentPresent = Nothing
+		, usesLocationLog = True
+		}
+	withKeyOptions (keyOptions o) False seeker
 		(\kai -> commandAction . startKey from i kai =<< getNumCopies)
-		(withFilesInGit $ commandAction . (whenAnnexed (start from i)))
-		=<< workTreeItems (fsckFiles o)
+		(withFilesInGitAnnex ww seeker)
+		=<< workTreeItems ww (fsckFiles o)
 	cleanupIncremental i
 	void $ tryIO $ recordActivity Fsck u
+  where
+	ww = WarnUnmatchLsFiles
 
 checkDeadRepo :: UUID -> Annex ()
 checkDeadRepo u =
 	whenM ((==) DeadTrusted <$> lookupTrust u) $
 		earlyWarning "Warning: Fscking a repository that is currently marked as dead."
 
-start :: Maybe Remote -> Incremental -> FilePath -> Key -> CommandStart
-start from inc file key = Backend.getBackend file key >>= \case
+start :: Maybe Remote -> Incremental -> SeekInput -> RawFilePath -> Key -> CommandStart
+start from inc si file key = Backend.getBackend (fromRawFilePath file) key >>= \case
 	Nothing -> stop
 	Just backend -> do
-		numcopies <- getFileNumCopies file
+		(numcopies, _mincopies) <- getFileNumMinCopies file
 		case from of
 			Nothing -> go $ perform key file backend numcopies
 			Just r -> go $ performRemote key afile backend numcopies r
   where
-	go = runFsck inc (mkActionItem (key, afile)) key
+	go = runFsck inc si (mkActionItem (key, afile)) key
 	afile = AssociatedFile (Just file)
 
-perform :: Key -> FilePath -> Backend -> NumCopies -> Annex Bool
+perform :: Key -> RawFilePath -> Backend -> NumCopies -> Annex Bool
 perform key file backend numcopies = do
 	keystatus <- getKeyFileStatus key file
 	check
 		-- order matters
 		[ fixLink key file
+		, fixObjectLocation key
 		, verifyLocationLog key keystatus ai
 		, verifyRequiredContent key ai
 		, verifyAssociatedFiles key keystatus file
@@ -148,46 +162,57 @@ performRemote key afile backend numcopies remote =
 	dispatch (Right True) = withtmp $ \tmpfile ->
 		getfile tmpfile >>= \case
 			Nothing -> go True Nothing
-			Just True -> go True (Just tmpfile)
-			Just False -> do
+			Just (Right verification) -> go True (Just (tmpfile, verification))
+			Just (Left _) -> do
 				warning "failed to download file from remote"
 				void $ go True Nothing
 				return False
 	dispatch (Right False) = go False Nothing
-	go present localcopy = check
+	go present lv = check
 		[ verifyLocationLogRemote key ai remote present
 		, verifyRequiredContent key ai
-		, withLocalCopy localcopy $ checkKeySizeRemote key remote ai
-		, withLocalCopy localcopy $ checkBackendRemote backend key remote ai
+		, withLocalCopy (fmap fst lv) $ checkKeySizeRemote key remote ai
+		, case fmap snd lv of
+			Just Verified -> return True
+			_ -> withLocalCopy (fmap fst lv) $
+				checkBackendRemote backend key remote ai
 		, checkKeyNumCopies key afile numcopies
 		]
 	ai = mkActionItem (key, afile)
 	withtmp a = do
+		-- Put it in the gitAnnexTmpObjectDir since that's on a
+		-- filesystem where object temp files are normally
+		-- stored. The pid prevents multiple fsck processes
+		-- contending over the same file. (Multiple threads cannot,
+		-- because OnlyActionOn is used.)
 		pid <- liftIO getPID
 		t <- fromRepo gitAnnexTmpObjectDir
 		createAnnexDirectory t
-		let tmp = t </> "fsck" ++ show pid ++ "." ++ keyFile key
-		let cleanup = liftIO $ catchIO (removeFile tmp) (const noop)
+		let tmp = t P.</> "fsck" <> toRawFilePath (show pid) <> "." <> keyFile key
+		let cleanup = liftIO $ catchIO (R.removeLink tmp) (const noop)
 		cleanup
 		cleanup `after` a tmp
-	getfile tmp = ifM (checkDiskSpace (Just (takeDirectory tmp)) key 0 True)
-		( ifM (Remote.retrieveKeyFileCheap remote key afile tmp)
-			( return (Just True)
+	getfile tmp = ifM (checkDiskSpace (Just (P.takeDirectory tmp)) key 0 True)
+		( ifM (getcheap tmp)
+			( return (Just (Right UnVerified))
 			, ifM (Annex.getState Annex.fast)
 				( return Nothing
-				, Just . fst <$>
-					Remote.retrieveKeyFile remote key (AssociatedFile Nothing) tmp dummymeter
+				, Just <$> tryNonAsync (getfile' tmp)
 				)
 			)
-		, return (Just False)
+		, return Nothing
 		)
+	getfile' tmp = Remote.retrieveKeyFile remote key (AssociatedFile Nothing) (fromRawFilePath tmp) dummymeter (RemoteVerify remote)
 	dummymeter _ = noop
+	getcheap tmp = case Remote.retrieveKeyFileCheap remote of
+		Just a -> isRight <$> tryNonAsync (a key afile (fromRawFilePath tmp))
+		Nothing -> return False
 
-startKey :: Maybe Remote -> Incremental -> (Key, ActionItem) -> NumCopies -> CommandStart
-startKey from inc (key, ai) numcopies =
-	case Backend.maybeLookupBackendVariety (keyVariety key) of
+startKey :: Maybe Remote -> Incremental -> (SeekInput, Key, ActionItem) -> NumCopies -> CommandStart
+startKey from inc (si, key, ai) numcopies =
+	Backend.maybeLookupBackendVariety (fromKey keyVariety key) >>= \case
 		Nothing -> stop
-		Just backend -> runFsck inc ai key $
+		Just backend -> runFsck inc si ai key $
 			case from of
 				Nothing -> performKey key backend numcopies
 				Just r -> performRemote key (AssociatedFile Nothing) backend numcopies r
@@ -206,7 +231,7 @@ check :: [Annex Bool] -> Annex Bool
 check cs = and <$> sequence cs
 
 {- Checks that symlinks points correctly to the annexed content. -}
-fixLink :: Key -> FilePath -> Annex Bool
+fixLink :: Key -> RawFilePath -> Annex Bool
 fixLink key file = do
 	want <- calcRepo $ gitAnnexLink file key
 	have <- getAnnexLinkTarget file
@@ -214,49 +239,97 @@ fixLink key file = do
 	return True
   where
 	go want have
-		| want /= fromInternalGitPath (fromRawFilePath have) = do
+		| want /= fromInternalGitPath have = do
 			showNote "fixing link"
-			liftIO $ createDirectoryIfMissing True (parentDir file)
-			liftIO $ removeFile file
+			createWorkTreeDirectory (parentDir file)
+			liftIO $ R.removeLink file
 			addAnnexLink want file
 		| otherwise = noop
 
+{- A repository that supports symlinks and is not bare may have in the past
+ - been bare, or not supported symlinks. If so, the object may be located
+ - in a directory other than the one where annex symlinks point to. Moves
+ - the object in that case.
+ -
+ - Also if a repository has been converted to bare, or moved to a crippled
+ - filesystem not supporting symlinks, the object file will be moved
+ - to the other location.
+ -}
+fixObjectLocation :: Key -> Annex Bool
+fixObjectLocation key = do
+#ifdef mingw32_HOST_OS
+	-- Windows does not allow locked files to be renamed, but annex
+	-- links are also not used on Windows.
+	return True
+#else
+	loc <- calcRepo (gitAnnexLocation key)
+	idealloc <- calcRepo (gitAnnexLocation' (const (pure True)) key)
+	if loc == idealloc
+		then return True
+		else ifM (liftIO $ R.doesPathExist loc)
+			( moveobjdir loc idealloc
+				`catchNonAsync` \_e -> return True
+			, return True
+			)
+  where
+	moveobjdir src dest = do
+		let srcdir = parentDir src
+		let destdir = parentDir dest
+		showNote "normalizing object location"
+		-- When the content file is moved, it will
+		-- appear to other processes as if it has been removed.
+		-- That should never happen to a process that has used
+		-- lockContentShared, so avoid it by locking the content
+		-- for removal, although it's not really being removed.
+		lockContentForRemoval key (return True) $ \_lck -> do
+			-- Thaw the content directory to allow renaming it.
+			thawContentDir src
+			createAnnexDirectory (parentDir destdir)
+			liftIO $ renameDirectory
+				(fromRawFilePath srcdir)
+				(fromRawFilePath destdir)
+			-- Since the directory was moved, lockContentForRemoval
+			-- will not be able to remove the lock file it
+			-- made. So, remove the lock file here.
+			mlockfile <- contentLockFile key =<< getVersion
+			liftIO $ maybe noop (removeWhenExistsWith R.removeLink) mlockfile
+			freezeContentDir dest
+			cleanObjectDirs src
+			return True
+#endif
+
 {- Checks that the location log reflects the current status of the key,
  - in this repository only. -}
 verifyLocationLog :: Key -> KeyStatus -> ActionItem -> Annex Bool
 verifyLocationLog key keystatus ai = do
-	direct <- isDirect
-	obj <- calcRepo $ gitAnnexLocation key
-	present <- if not direct && isKeyUnlockedThin keystatus
-		then liftIO (doesFileExist obj)
+	obj <- calcRepo (gitAnnexLocation key)
+	present <- if isKeyUnlockedThin keystatus
+		then liftIO (doesFileExist (fromRawFilePath obj))
 		else inAnnex key
 	u <- getUUID
 	
 	{- Since we're checking that a key's object file is present, throw
 	 - in a permission fixup here too. -}
-	when (present && not direct) $ do
+	when present $ do
 		void $ tryIO $ case keystatus of
 			KeyUnlockedThin -> thawContent obj
 			KeyLockedThin -> thawContent obj
 			_ -> freezeContent obj
-		unlessM (isContentWritePermOk obj) $
-			warning $ "** Unable to set correct write mode for " ++ obj ++ " ; perhaps you don't own that file"
-	whenM (liftIO $ doesDirectoryExist $ parentDir obj) $
+		checkContentWritePerm obj >>= \case
+			Nothing -> warning $ "** Unable to set correct write mode for " ++ fromRawFilePath obj ++ " ; perhaps you don't own that file, or perhaps it has an xattr or ACL set"
+			_ -> return ()
+	whenM (liftIO $ R.doesPathExist $ parentDir obj) $
 		freezeContentDir obj
 
 	{- Warn when annex.securehashesonly is set and content using an 
 	 - insecure hash is present. This should only be able to happen
 	 - if the repository already contained the content before the
 	 - config was set. -}
-	when (present && not (cryptographicallySecure (keyVariety key))) $
+	whenM (pure present <&&> (not <$> Backend.isCryptographicallySecure key)) $
 		whenM (annexSecureHashesOnly <$> Annex.getGitConfig) $
-			warning $ "** Despite annex.securehashesonly being set, " ++ obj ++ " has content present in the annex using an insecure " ++ decodeBS (formatKeyVariety (keyVariety key)) ++ " key"
+			warning $ "** Despite annex.securehashesonly being set, " ++ fromRawFilePath obj ++ " has content present in the annex using an insecure " ++ decodeBS (formatKeyVariety (fromKey keyVariety key)) ++ " key"
 
-	{- In direct mode, modified files will show up as not present,
-	 - but that is expected and not something to do anything about. -}
-	if direct && not present
-		then return True
-		else verifyLocationLog' key ai present u (logChange key u)
+	verifyLocationLog' key ai present u (logChange key u)
 
 verifyLocationLogRemote :: Key -> ActionItem -> Remote -> Bool -> Annex Bool
 verifyLocationLogRemote key ai remote present =
@@ -275,7 +348,7 @@ verifyLocationLog' key ai present u updatestatus = do
 			fix InfoMissing
 			warning $
 				"** Based on the location log, " ++
-				actionItemDesc ai ++
+				decodeBS (actionItemDesc ai) ++
 				"\n** was expected to be present, " ++
 				"but its content is missing."
 			return False
@@ -295,100 +368,88 @@ verifyLocationLog' key ai present u updatestatus = do
 {- Verifies that all repos that are required to contain the content do,
  - checking against the location log. -}
 verifyRequiredContent :: Key -> ActionItem -> Annex Bool
-verifyRequiredContent key ai@(ActionItemAssociatedFile afile _) = do
-	requiredlocs <- S.fromList . M.keys <$> requiredContentMap
-	if S.null requiredlocs
-		then return True
-		else do
-			presentlocs <- S.fromList <$> loggedLocations key
-			missinglocs <- filterM
-				(\u -> isRequiredContent (Just u) S.empty (Just key) afile False)
-				(S.toList $ S.difference requiredlocs presentlocs)
-			if null missinglocs
-				then return True
-				else do
-					missingrequired <- Remote.prettyPrintUUIDs "missingrequired" missinglocs
-					warning $
-						"** Required content " ++
-						actionItemDesc ai ++
-						" is missing from these repositories:\n" ++
-						missingrequired
-					return False
+verifyRequiredContent key ai@(ActionItemAssociatedFile afile _) = case afile of
+	-- Can't be checked if there's no associated file.
+	AssociatedFile Nothing -> return True
+	AssociatedFile (Just _) -> do
+		requiredlocs <- S.fromList . M.keys <$> requiredContentMap
+		if S.null requiredlocs
+			then return True
+			else go requiredlocs
+  where
+	go requiredlocs = do
+		presentlocs <- S.fromList <$> loggedLocations key
+		missinglocs <- filterM
+			(\u -> isRequiredContent (Just u) S.empty (Just key) afile False)
+			(S.toList $ S.difference requiredlocs presentlocs)
+		if null missinglocs
+			then return True
+			else do
+				missingrequired <- Remote.prettyPrintUUIDs "missingrequired" missinglocs
+				warning $
+					"** Required content " ++
+					decodeBS (actionItemDesc ai) ++
+					" is missing from these repositories:\n" ++
+					missingrequired
+				return False
 verifyRequiredContent _ _ = return True
 
 {- Verifies the associated file records. -}
-verifyAssociatedFiles :: Key -> KeyStatus -> FilePath -> Annex Bool
+verifyAssociatedFiles :: Key -> KeyStatus -> RawFilePath -> Annex Bool
 verifyAssociatedFiles key keystatus file = do
-	ifM isDirect (godirect, goindirect)
-	return True
-  where
-	godirect = do
-		fs <- Direct.addAssociatedFile key file
-		forM_ fs $ \f -> 
-			unlessM (liftIO $ doesFileExist f) $
-				void $ Direct.removeAssociatedFile key f
-	goindirect = when (isKeyUnlockedThin keystatus) $ do
+	when (isKeyUnlockedThin keystatus) $ do
 		f <- inRepo $ toTopFilePath file
 		afs <- Database.Keys.getAssociatedFiles key
 		unless (getTopFilePath f `elem` map getTopFilePath afs) $
 			Database.Keys.addAssociatedFile key f
-
-verifyWorkTree :: Key -> FilePath -> Annex Bool
-verifyWorkTree key file = do
-	ifM isDirect ( godirect, goindirect )
 	return True
-  where
-	{- Ensures that files whose content is available are in direct mode. -}
-	godirect = whenM (isJust <$> isAnnexLink file) $ do
-		v <- toDirectGen key file
-		case v of
-			Nothing -> noop
-			Just a -> do
-				showNote "fixing direct mode"
-				a
+
+verifyWorkTree :: Key -> RawFilePath -> Annex Bool
+verifyWorkTree key file = do
 	{- Make sure that a pointer file is replaced with its content,
 	 - when the content is available. -}
-	goindirect = do
-		mk <- liftIO $ isPointerFile file
-		case mk of
-			Just k | k == key -> whenM (inAnnex key) $ do
-				showNote "fixing worktree content"
-				replaceFile file $ \tmp -> do
-					mode <- liftIO $ catchMaybeIO $ fileMode <$> getFileStatus file
-					ifM (annexThin <$> Annex.getGitConfig)
-						( void $ linkFromAnnex key tmp mode
-						, do
-							obj <- calcRepo $ gitAnnexLocation key
-							void $ checkedCopyFile key obj tmp mode
-							thawContent tmp
-						)
-				Database.Keys.storeInodeCaches key [file]
-			_ -> return ()
+	mk <- liftIO $ isPointerFile file
+	case mk of
+		Just k | k == key -> whenM (inAnnex key) $ do
+			showNote "fixing worktree content"
+			replaceWorkTreeFile (fromRawFilePath file) $ \tmp -> do
+				let tmp' = toRawFilePath tmp
+				mode <- liftIO $ catchMaybeIO $ fileMode <$> R.getFileStatus file
+				ifM (annexThin <$> Annex.getGitConfig)
+					( void $ linkFromAnnex' key tmp' mode
+					, do
+						obj <- calcRepo (gitAnnexLocation key)
+						void $ checkedCopyFile key obj tmp' mode
+						thawContent tmp'
+					)
+				Database.Keys.storeInodeCaches key [tmp']
+		_ -> return ()
+	return True
 
 {- The size of the data for a key is checked against the size encoded in
  - the key's metadata, if available.
  -
- - Not checked when a file is unlocked, or in direct mode.
+ - Not checked when a file is unlocked.
  -}
 checkKeySize :: Key -> KeyStatus -> ActionItem -> Annex Bool
 checkKeySize _ KeyUnlockedThin _ = return True
 checkKeySize key _ ai = do
 	file <- calcRepo $ gitAnnexLocation key
-	ifM (liftIO $ doesFileExist file)
+	ifM (liftIO $ R.doesPathExist file)
 		( checkKeySizeOr badContent key file ai
 		, return True
 		)
 
-withLocalCopy :: Maybe FilePath -> (FilePath -> Annex Bool) -> Annex Bool
+withLocalCopy :: Maybe RawFilePath -> (RawFilePath -> Annex Bool) -> Annex Bool
 withLocalCopy Nothing _ = return True
 withLocalCopy (Just localcopy) f = f localcopy
 
-checkKeySizeRemote :: Key -> Remote -> ActionItem -> FilePath -> Annex Bool
+checkKeySizeRemote :: Key -> Remote -> ActionItem -> RawFilePath -> Annex Bool
 checkKeySizeRemote key remote ai localcopy =
 	checkKeySizeOr (badContentRemote remote localcopy) key localcopy ai
 
-checkKeySizeOr :: (Key -> Annex String) -> Key -> FilePath -> ActionItem -> Annex Bool
-checkKeySizeOr bad key file ai = case keySize key of
+checkKeySizeOr :: (Key -> Annex String) -> Key -> RawFilePath -> ActionItem -> Annex Bool
+checkKeySizeOr bad key file ai = case fromKey keySize key of
 	Nothing -> return True
 	Just size -> do
 		size' <- liftIO $ getFileSize file
@@ -401,7 +462,7 @@ checkKeySizeOr bad key file ai = case keySize key of
 	badsize a b = do
 		msg <- bad key
 		warning $ concat
-			[ actionItemDesc ai
+			[ decodeBS (actionItemDesc ai)
 			, ": Bad file size ("
 			, compareSizes storageUnits True a b
 			, "); "
@@ -419,11 +480,11 @@ checkKeyUpgrade backend key ai (AssociatedFile (Just file)) =
 	case Types.Backend.canUpgradeKey backend of
 		Just a | a key -> do
 			warning $ concat
-				[ actionItemDesc ai
+				[ decodeBS (actionItemDesc ai)
 				, ": Can be upgraded to an improved key format. "
 				, "You can do so by running: git annex migrate --backend="
-				, decodeBS (formatKeyVariety (keyVariety key)) ++ " "
-				, file
+				, decodeBS (formatKeyVariety (fromKey keyVariety key)) ++ " "
+				, decodeBS file
 				]
 			return True
 		_ -> return True
@@ -435,76 +496,84 @@ checkKeyUpgrade _ _ _ (AssociatedFile Nothing) =
 
 {- Runs the backend specific check on a key's content object.
  -
- - When a file is unlocked, it may be a hard link to the object,
- - thus when the user modifies the file, the object will be modified and
+ - When a annex.this is set, an unlocked file may be a hard link to the object.
+ - Thus when the user modifies the file, the object will be modified and
  - not pass the check, and we don't want to find an error in this case.
- - So, skip the check if the key is unlocked and modified.
- -
- - In direct mode this is not done if the file has clearly been modified,
- - because modification of direct mode files is allowed. It's still done
- - if the file does not appear modified, to catch disk corruption, etc.
  -}
 checkBackend :: Backend -> Key -> KeyStatus -> AssociatedFile -> Annex Bool
-checkBackend backend key keystatus afile = go =<< isDirect
-  where
-	go False = do
-		content <- calcRepo $ gitAnnexLocation key
-		ifM (pure (isKeyUnlockedThin keystatus) <&&> (not <$> isUnmodified key content))
-			( nocheck
-			, checkBackendOr badContent backend key content ai
-			)
-	go True = case afile of
-		AssociatedFile Nothing -> nocheck
-		AssociatedFile (Just f) -> checkdirect f
-	checkdirect file = ifM (Direct.goodContent key file)
-		( checkBackendOr' (badContentDirect file) backend key file ai
-			(Direct.goodContent key file)
-		, nocheck
+checkBackend backend key keystatus afile = do
+	content <- calcRepo (gitAnnexLocation key)
+	ifM (pure (isKeyUnlockedThin keystatus) <&&> (not <$> isUnmodified key content))
+		( nocheck
+		, do
+			mic <- withTSDelta (liftIO . genInodeCache content)
+			ifM (checkBackendOr badContent backend key content ai)
+				( do
+					checkInodeCache key content mic ai
+					return True
+				, return False
+				)
 		)
+  where
 	nocheck = return True
 
 	ai = mkActionItem (key, afile)
 
-checkBackendRemote :: Backend -> Key -> Remote -> ActionItem -> FilePath -> Annex Bool
+checkBackendRemote :: Backend -> Key -> Remote -> ActionItem -> RawFilePath -> Annex Bool
 checkBackendRemote backend key remote ai localcopy =
 	checkBackendOr (badContentRemote remote localcopy) backend key localcopy ai
 
-checkBackendOr :: (Key -> Annex String) -> Backend -> Key -> FilePath -> ActionItem -> Annex Bool
+checkBackendOr :: (Key -> Annex String) -> Backend -> Key -> RawFilePath -> ActionItem -> Annex Bool
 checkBackendOr bad backend key file ai =
-	checkBackendOr' bad backend key file ai (return True)
-
--- The postcheck action is run after the content is verified,
--- in order to detect situations where the file is changed while being
--- verified (particularly in direct mode).
-checkBackendOr' :: (Key -> Annex String) -> Backend -> Key -> FilePath -> ActionItem -> Annex Bool -> Annex Bool
-checkBackendOr' bad backend key file ai postcheck =
 	case Types.Backend.verifyKeyContent backend of
-		Nothing -> return True
 		Just verifier -> do
 			ok <- verifier key file
-			ifM postcheck
-				( do
-					unless ok $ do
-						msg <- bad key
+			unless ok $ do
+				msg <- bad key
+				warning $ concat
+					[ decodeBS (actionItemDesc ai)
+					, ": Bad file content; "
+					, msg
+					]
+			return ok
+		Nothing -> return True
+
+{- Check, if there are InodeCaches recorded for a key, that one of them
+ - matches the object file. There are situations where the InodeCache
+ - of the object file does not get recorded, including a v8 upgrade.
+ - There may also be situations where the wrong InodeCache is recorded,
+ - if inodes are not stable.
+ -
+ - This must be called after the content of the object file has been
+ - verified to be correct. The InodeCache is generated again to detect if
+ - the object file was changed while the content was being verified.
+ -}
+checkInodeCache :: Key -> RawFilePath -> Maybe InodeCache -> ActionItem -> Annex ()
+checkInodeCache key content mic ai = case mic of
+	Nothing -> noop
+	Just ic -> do
+		ics <- Database.Keys.getInodeCaches key
+		unless (null ics) $
+			unlessM (isUnmodifiedCheapLowLevel ic ics) $ do
+				withTSDelta (liftIO . genInodeCache content) >>= \case
+					Nothing -> noop
+					Just ic' -> whenM (compareInodeCaches ic ic') $ do
 						warning $ concat
-							[ actionItemDesc ai
-							, ": Bad file content; "
-							, msg
+							[ decodeBS (actionItemDesc ai)
+							, ": Stale or missing inode cache; updating."
 							]
-					return ok
-				, return True
-				)
+						Database.Keys.addInodeCaches key [ic]
 
 checkKeyNumCopies :: Key -> AssociatedFile -> NumCopies -> Annex Bool
 checkKeyNumCopies key afile numcopies = do
 	let (desc, hasafile) = case afile of
 		AssociatedFile Nothing -> (serializeKey key, False)
-		AssociatedFile (Just af) -> (af, True)
+		AssociatedFile (Just af) -> (fromRawFilePath af, True)
 	locs <- loggedLocations key
 	(untrustedlocations, otherlocations) <- trustPartition UnTrusted locs
 	(deadlocations, safelocations) <- trustPartition DeadTrusted otherlocations
-	let present = NumCopies (length safelocations)
-	if present < numcopies
+	let present = length safelocations
+	if present < fromNumCopies numcopies
 		then ifM (pure (not hasafile) <&&> checkDead key)
 			( do
 				showLongNote $ "This key is dead, skipping."
@@ -513,21 +582,21 @@ checkKeyNumCopies key afile numcopies = do
 				untrusted <- Remote.prettyPrintUUIDs "untrusted" untrustedlocations
 				dead <- Remote.prettyPrintUUIDs "dead" deadlocations
 				warning $ missingNote desc present numcopies untrusted dead
-				when (fromNumCopies present == 0 && not hasafile) $
+				when (present == 0 && not hasafile) $
 					showLongNote "(Avoid this check by running: git annex dead --key )"
 				return False
 			)
 		else return True
 
-missingNote :: String -> NumCopies -> NumCopies -> String -> String -> String
-missingNote file (NumCopies 0) _ [] dead = 
+missingNote :: String -> Int -> NumCopies -> String -> String -> String
+missingNote file 0 _ [] dead = 
 		"** No known copies exist of " ++ file ++ honorDead dead
-missingNote file (NumCopies 0) _ untrusted dead =
+missingNote file 0 _ untrusted dead =
 		"Only these untrusted locations may have copies of " ++ file ++
 		"\n" ++ untrusted ++
 		"Back it up to trusted locations with git-annex copy." ++ honorDead dead
 missingNote file present needed [] _ =
-		"Only " ++ show (fromNumCopies present) ++ " of " ++ show (fromNumCopies needed) ++ 
+		"Only " ++ show present ++ " of " ++ show (fromNumCopies needed) ++ 
 		" trustworthy copies exist of " ++ file ++
 		"\nBack it up with git-annex copy."
 missingNote file present needed untrusted dead = 
@@ -544,50 +613,43 @@ honorDead dead
 badContent :: Key -> Annex String
 badContent key = do
 	dest <- moveBad key
-	return $ "moved to " ++ dest
-
-{- Bad content is left where it is, but we touch the file, so it'll be
- - committed to a new key. -}
-badContentDirect :: FilePath -> Key -> Annex String
-badContentDirect file key = do
-	void $ liftIO $ catchMaybeIO $ touchFile file
-	logStatus key InfoMissing
-	return "left in place for you to examine"
+	return $ "moved to " ++ fromRawFilePath dest
 
 {- Bad content is dropped from the remote. We have downloaded a copy
  - from the remote to a temp file already (in some cases, it's just a
  - symlink to a file in the remote). To avoid any further data loss,
  - that temp file is moved to the bad content directory unless 
  - the local annex has a copy of the content. -}
-badContentRemote :: Remote -> FilePath -> Key -> Annex String
+badContentRemote :: Remote -> RawFilePath -> Key -> Annex String
 badContentRemote remote localcopy key = do
 	bad <- fromRepo gitAnnexBadDir
-	let destbad = bad </> keyFile key
-	movedbad <- ifM (inAnnex key <||> liftIO (doesFileExist destbad))
+	let destbad = bad P.</> keyFile key
+	let destbad' = fromRawFilePath destbad
+	movedbad <- ifM (inAnnex key <||> liftIO (doesFileExist destbad'))
 		( return False
 		, do
 			createAnnexDirectory (parentDir destbad)
 			liftIO $ catchDefaultIO False $
-				ifM (isSymbolicLink <$> getSymbolicLinkStatus localcopy)
-					( copyFileExternal CopyTimeStamps localcopy destbad
+				ifM (isSymbolicLink <$> R.getSymbolicLinkStatus localcopy)
+					( copyFileExternal CopyTimeStamps (fromRawFilePath localcopy) destbad'
 					, do
 						moveFile localcopy destbad
 						return True
 					)
 		)
 
-	dropped <- Remote.removeKey remote key
-	when dropped $
+	dropped <- tryNonAsync (Remote.removeKey remote key)
+	when (isRight dropped) $
 		Remote.logStatus remote key InfoMissing
 	return $ case (movedbad, dropped) of
-		(True, True) -> "moved from " ++ Remote.name remote ++
-			" to " ++ destbad
-		(False, True) -> "dropped from " ++ Remote.name remote
-		(_, False) -> "failed to drop from" ++ Remote.name remote
+		(True, Right ()) -> "moved from " ++ Remote.name remote ++
+			" to " ++ fromRawFilePath destbad
+		(False, Right ()) -> "dropped from " ++ Remote.name remote
+		(_, Left e) -> "failed to drop from" ++ Remote.name remote ++ ": " ++ show e
 
-runFsck :: Incremental -> ActionItem -> Key -> Annex Bool -> CommandStart
-runFsck inc ai key a = stopUnless (needFsck inc key) $
-	starting "fsck" ai $ do
+runFsck :: Incremental -> SeekInput -> ActionItem -> Key -> Annex Bool -> CommandStart
+runFsck inc si ai key a = stopUnless (needFsck inc key) $
+	starting "fsck" (OnlyActionOn key ai) si $ do
 		ok <- a
 		when ok $
 			recordFsckTime inc key
@@ -614,10 +676,10 @@ recordStartTime :: UUID -> Annex ()
 recordStartTime u = do
 	f <- fromRepo (gitAnnexFsckState u)
 	createAnnexDirectory $ parentDir f
-	liftIO $ nukeFile f
-	liftIO $ withFile f WriteMode $ \h -> do
+	liftIO $ removeWhenExistsWith R.removeLink f
+	liftIO $ withFile (fromRawFilePath f) WriteMode $ \h -> do
 #ifndef mingw32_HOST_OS
-		t <- modificationTime <$> getFileStatus f
+		t <- modificationTime <$> R.getFileStatus f
 #else
 		t <- getPOSIXTime
 #endif
@@ -628,16 +690,17 @@ recordStartTime u = do
 	showTime = show
 
 resetStartTime :: UUID -> Annex ()
-resetStartTime u = liftIO . nukeFile =<< fromRepo (gitAnnexFsckState u)
+resetStartTime u = liftIO . removeWhenExistsWith R.removeLink
+	=<< fromRepo (gitAnnexFsckState u)
 
 {- Gets the incremental fsck start time. -}
 getStartTime :: UUID -> Annex (Maybe EpochTime)
 getStartTime u = do
 	f <- fromRepo (gitAnnexFsckState u)
 	liftIO $ catchDefaultIO Nothing $ do
-		timestamp <- modificationTime <$> getFileStatus f
+		timestamp <- modificationTime <$> R.getFileStatus f
 		let fromstatus = Just (realToFrac timestamp)
-		fromfile <- parsePOSIXTime <$> readFile f
+		fromfile <- parsePOSIXTime <$> readFile (fromRawFilePath f)
 		return $ if matchingtimestamp fromfile fromstatus
 			then Just timestamp
 			else Nothing
@@ -687,7 +750,7 @@ cleanupIncremental _ = return ()
 openFsckDb :: UUID -> Annex FsckDb.FsckHandle
 openFsckDb u = do
 	h <- FsckDb.openDb u
-	Annex.addCleanup FsckCleanup $
+	Annex.addCleanupAction FsckCleanup $
 		FsckDb.closeDb h
 	return h
 
@@ -697,42 +760,3 @@ withFsckDb (StartIncremental h) a = a h
 withFsckDb NonIncremental _ = noop
 withFsckDb (ScheduleIncremental _ _ i) a = withFsckDb i a
 
-data KeyStatus
-	= KeyMissing
-	| KeyPresent
-	| KeyUnlockedThin
-	-- ^ An annex.thin worktree file is hard linked to the object.
-	| KeyLockedThin
-	-- ^ The object has hard links, but the file being fscked
-	-- is not the one that hard links to it.
-	deriving (Show)
-
-isKeyUnlockedThin :: KeyStatus -> Bool
-isKeyUnlockedThin KeyUnlockedThin = True
-isKeyUnlockedThin KeyLockedThin = False
-isKeyUnlockedThin KeyPresent = False
-isKeyUnlockedThin KeyMissing = False
-
-getKeyStatus :: Key -> Annex KeyStatus
-getKeyStatus key = ifM isDirect
-	( return KeyUnlockedThin
-	, catchDefaultIO KeyMissing $ do
-		afs <- not . null <$> Database.Keys.getAssociatedFiles key
-		obj <- calcRepo $ gitAnnexLocation key
-		multilink <- ((> 1) . linkCount <$> liftIO (getFileStatus obj))
-		return $ if multilink && afs
-			then KeyUnlockedThin
-			else KeyPresent
-	)
-
-getKeyFileStatus :: Key -> FilePath -> Annex KeyStatus
-getKeyFileStatus key file = do
-	s <- getKeyStatus key
-	case s of
-		KeyUnlockedThin -> catchDefaultIO KeyUnlockedThin $
-			ifM (isJust <$> isAnnexLink file)
-				( return KeyLockedThin
-				, return KeyUnlockedThin
-				)
-		_ -> return s
-
