@@ -1,17 +1,18 @@
 {- git-annex command
  -
- - Copyright 2011-2016 Joey Hess <id@joeyh.name>
+ - Copyright 2011-2022 Joey Hess <id@joeyh.name>
  -
  - Licensed under the GNU AGPL version 3 or higher.
  -}
 
-{-# LANGUAGE BangPatterns, DeriveDataTypeable #-}
+{-# LANGUAGE BangPatterns, DeriveDataTypeable, PackageImports, OverloadedStrings #-}
 
 module Command.Info where
 
 import "mtl" Control.Monad.State.Strict
 import qualified Data.Map.Strict as M
 import qualified Data.Vector as V
+import qualified System.FilePath.ByteString as P
 import Data.Ord
 import qualified Data.Semigroup as Sem
 import Prelude
@@ -21,6 +22,7 @@ import qualified Git
 import qualified Annex
 import qualified Remote
 import qualified Types.Remote as Remote
+import qualified Annex.SpecialRemote as SpecialRemote
 import Utility.DataUnits
 import Utility.DiskFree
 import Annex.Content
@@ -30,8 +32,6 @@ import Logs.UUID
 import Logs.Trust
 import Logs.Location
 import Annex.NumCopies
-import Remote
-import Config
 import Git.Config (boolConfig)
 import qualified Git.LsTree as LsTree
 import Utility.Percentage
@@ -51,23 +51,23 @@ import qualified Command.Unused
 type Stat = StatState (Maybe (String, StatState String))
 
 -- data about a set of keys
-data KeyData = KeyData
+data KeyInfo = KeyInfo
 	{ countKeys :: Integer
 	, sizeKeys :: Integer
 	, unknownSizeKeys :: Integer
 	, backendsKeys :: M.Map KeyVariety Integer
 	}
 	
-instance Sem.Semigroup KeyData where
-	a <> b = KeyData
+instance Sem.Semigroup KeyInfo where
+	a <> b = KeyInfo
 		{ countKeys = countKeys a + countKeys b
 		, sizeKeys = sizeKeys a + sizeKeys b
 		, unknownSizeKeys = unknownSizeKeys a + unknownSizeKeys b
 		, backendsKeys = backendsKeys a <> backendsKeys b
 		}
 
-instance Monoid KeyData where
-	mempty = KeyData 0 0 0 M.empty
+instance Monoid KeyInfo where
+	mempty = KeyInfo 0 0 0 M.empty
 
 data NumCopiesStats = NumCopiesStats
 	{ numCopiesVarianceMap :: M.Map Variance Integer
@@ -83,9 +83,9 @@ instance Show Variance where
 
 -- cached info that multiple Stats use
 data StatInfo = StatInfo
-	{ presentData :: Maybe KeyData
-	, referencedData :: Maybe KeyData
-	, repoData :: M.Map UUID KeyData
+	{ presentData :: Maybe KeyInfo
+	, referencedData :: Maybe KeyInfo
+	, repoData :: M.Map UUID KeyInfo
 	, numCopiesStats :: Maybe NumCopiesStats
 	, infoOptions :: InfoOptions
 	}
@@ -106,6 +106,7 @@ data InfoOptions = InfoOptions
 	{ infoFor :: CmdParams
 	, bytesOption :: Bool
 	, batchOption :: BatchMode
+	, autoenableOption :: Bool
 	}
 
 optParser :: CmdParamsDesc -> Parser InfoOptions
@@ -115,19 +116,26 @@ optParser desc = InfoOptions
 		( long "bytes"
 		<> help "display file sizes in bytes"
 		)
-	<*> parseBatchOption
+	<*> parseBatchOption False
+	<*> switch
+		( long "autoenable"
+		<> help "list special remotes that are configured to autoenable"
+		)
 
 seek :: InfoOptions -> CommandSeek
 seek o = case batchOption o of
 	NoBatch -> withWords (commandAction . start o) (infoFor o)
-	Batch fmt -> batchInput fmt Right (itemInfo o)
+	Batch fmt -> batchOnly Nothing (infoFor o) $
+		batchInput fmt (pure . Right) (itemInfo o)
 
 start :: InfoOptions -> [String] -> CommandStart
 start o [] = do
-	globalInfo o
+	if autoenableOption o
+		then autoenableInfo
+		else globalInfo o
 	stop
 start o ps = do
-	mapM_ (itemInfo o) ps
+	mapM_ (\p -> itemInfo o (SeekInput [p], p)) ps
 	stop
 
 globalInfo :: InfoOptions -> Annex ()
@@ -136,36 +144,51 @@ globalInfo o = do
 	whenM ((==) DeadTrusted <$> lookupTrust u) $
 		earlyWarning "Warning: This repository is currently marked as dead."
 	stats <- selStats global_fast_stats global_slow_stats
-	showCustom "info" $ do
+	showCustom "info" (SeekInput []) $ do
 		evalStateT (mapM_ showStat stats) (emptyStatInfo o)
 		return True
 
-itemInfo :: InfoOptions -> String -> Annex ()
-itemInfo o p = ifM (isdir p)
-	( dirInfo o p
+autoenableInfo :: Annex ()
+autoenableInfo = showCustom "info" (SeekInput []) $ do
+	m <- SpecialRemote.specialRemoteNameMap
+		<$> SpecialRemote.autoEnableable
+	descm <- M.unionWith Remote.addName
+		<$> uuidDescMap
+		<*> pure (M.map toUUIDDesc m)
+	s <- Remote.prettyPrintUUIDsDescs
+		"autoenable special remotes"
+		descm (M.keys m)
+	showRaw (encodeBS s)
+	return True
+
+itemInfo :: InfoOptions -> (SeekInput, String) -> Annex ()
+itemInfo o (si, p) = ifM (isdir p)
+	( dirInfo o p si
 	, do
 		v <- Remote.byName' p
 		case v of
-			Right r -> remoteInfo o r
+			Right r -> remoteInfo o r si
 			Left _ -> do
 				v' <- Remote.nameToUUID' p
 				case v' of
-					Right u -> uuidInfo o u
-					Left _ -> ifAnnexed p 
-						(fileInfo o p)
-						(treeishInfo o p)
+					Right u -> uuidInfo o u si
+					Left _ -> do
+						relp <- liftIO $ relPathCwdToFile (toRawFilePath p)
+						ifAnnexed relp
+							(fileInfo o (fromRawFilePath relp) si)
+							(treeishInfo o p si)
 	)
   where
 	isdir = liftIO . catchBoolIO . (isDirectory <$$> getFileStatus)
 
-noInfo :: String -> Annex ()
-noInfo s = do
-	showStart "info" s
+noInfo :: String -> SeekInput -> Annex ()
+noInfo s si = do
+	showStart "info" (encodeBS s) si
 	showNote $ "not a directory or an annexed file or a treeish or a remote or a uuid"
 	showEndFail
 
-dirInfo :: InfoOptions -> FilePath -> Annex ()
-dirInfo o dir = showCustom (unwords ["info", dir]) $ do
+dirInfo :: InfoOptions -> FilePath -> SeekInput -> Annex ()
+dirInfo o dir si = showCustom (unwords ["info", dir]) si $ do
 	stats <- selStats
 		(tostats (dir_name:tree_fast_stats True))
 		(tostats tree_slow_stats)
@@ -174,12 +197,12 @@ dirInfo o dir = showCustom (unwords ["info", dir]) $ do
   where
 	tostats = map (\s -> s dir)
 
-treeishInfo :: InfoOptions -> String -> Annex ()
-treeishInfo o t = do
-	mi <- getTreeStatInfo o (Git.Ref t)
+treeishInfo :: InfoOptions -> String -> SeekInput -> Annex ()
+treeishInfo o t si = do
+	mi <- getTreeStatInfo o (Git.Ref (encodeBS t))
 	case mi of
-		Nothing -> noInfo t
-		Just i -> showCustom (unwords ["info", t]) $ do
+		Nothing -> noInfo t si
+		Just i -> showCustom (unwords ["info", t]) si $ do
 			stats <- selStats 
 				(tostats (tree_name:tree_fast_stats False)) 
 				(tostats tree_slow_stats)
@@ -188,13 +211,17 @@ treeishInfo o t = do
   where
 	tostats = map (\s -> s t)
 
-fileInfo :: InfoOptions -> FilePath -> Key -> Annex ()
-fileInfo o file k = showCustom (unwords ["info", file]) $ do
-	evalStateT (mapM_ showStat (file_stats file k)) (emptyStatInfo o)
-	return True
+fileInfo :: InfoOptions -> FilePath -> SeekInput -> Key -> Annex ()
+fileInfo o file si k = do
+	matcher <- Limit.getMatcher
+	let file' = toRawFilePath file
+	whenM (matcher $ MatchingFile $ FileInfo file' file' (Just k)) $
+		showCustom (unwords ["info", file]) si $ do
+			evalStateT (mapM_ showStat (file_stats file k)) (emptyStatInfo o)
+			return True
 
-remoteInfo :: InfoOptions -> Remote -> Annex ()
-remoteInfo o r = showCustom (unwords ["info", Remote.name r]) $ do
+remoteInfo :: InfoOptions -> Remote -> SeekInput -> Annex ()
+remoteInfo o r si = showCustom (unwords ["info", Remote.name r]) si $ do
 	i <- map (\(k, v) -> simpleStat k (pure v)) <$> Remote.getInfo r
 	let u = Remote.uuid r
 	l <- selStats 
@@ -203,8 +230,8 @@ remoteInfo o r = showCustom (unwords ["info", Remote.name r]) $ do
 	evalStateT (mapM_ showStat l) (emptyStatInfo o)
 	return True
 
-uuidInfo :: InfoOptions -> UUID -> Annex ()
-uuidInfo o u = showCustom (unwords ["info", fromUUID u]) $ do
+uuidInfo :: InfoOptions -> UUID -> SeekInput -> Annex ()
+uuidInfo o u si = showCustom (unwords ["info", fromUUID u]) si $ do
 	l <- selStats (uuid_fast_stats u) (uuid_slow_stats u)
 	evalStateT (mapM_ showStat l) (emptyStatInfo o)
 	return True
@@ -221,8 +248,7 @@ selStats fast_stats slow_stats = do
  -}
 global_fast_stats :: [Stat]
 global_fast_stats = 
-	[ repository_mode
-	, repo_list Trusted
+	[ repo_list Trusted
 	, repo_list SemiTrusted
 	, repo_list UnTrusted
 	, transfer_list
@@ -267,7 +293,6 @@ file_stats f k =
 remote_fast_stats :: Remote -> [Stat]
 remote_fast_stats r = map (\s -> s r)
 	[ remote_name
-	, remote_trust
 	, remote_cost
 	, remote_type
 	]
@@ -276,6 +301,7 @@ uuid_fast_stats :: UUID -> [Stat]
 uuid_fast_stats u = map (\s -> s u)
 	[ repo_uuid
 	, repo_description
+	, repo_trust
 	]
 
 uuid_slow_stats :: UUID -> [Stat]
@@ -307,27 +333,17 @@ showStat :: Stat -> StatState ()
 showStat s = maybe noop calc =<< s
   where
 	calc (desc, a) = do
-		(lift . showHeader) desc
-		lift . showRaw =<< a
-
-repository_mode :: Stat
-repository_mode = simpleStat "repository mode" $ lift $
-	ifM isDirect 
-		( return "direct"
-		, ifM (fromRepo Git.repoIsLocalBare)
-			( return "bare"
-			, return "indirect"
-			)
-		)
+		(lift . showHeader . encodeBS) desc
+		lift . showRaw . encodeBS =<< a
 
 repo_list :: TrustLevel -> Stat
 repo_list level = stat n $ nojson $ lift $ do
 	us <- filter (/= NoUUID) . M.keys 
-		<$> (M.union <$> (M.map fromUUIDDesc <$> uuidDescMap) <*> remoteMap Remote.name)
+		<$> (M.union <$> (M.map fromUUIDDesc <$> uuidDescMap) <*> Remote.remoteMap Remote.name)
 	rs <- fst <$> trustPartition level us
 	countRepoList (length rs)
 		-- This also handles json display.
-		<$> prettyPrintUUIDs n rs
+		<$> Remote.prettyPrintUUIDs n rs
   where
 	n = showTrustLevel level ++ " repositories"
 
@@ -353,9 +369,8 @@ repo_description = simpleStat "description" . lift . Remote.prettyUUID
 repo_uuid :: UUID -> Stat
 repo_uuid = simpleStat "uuid" . pure . fromUUID
 
-remote_trust :: Remote -> Stat
-remote_trust r = simpleStat "trust" $ lift $
-	showTrustLevel <$> lookupTrust (Remote.uuid r)
+repo_trust :: UUID -> Stat
+repo_trust u = simpleStat "trust" $ lift $ showTrustLevel <$> lookupTrust u
 
 remote_cost :: Remote -> Stat
 remote_cost r = simpleStat "cost" $ pure $
@@ -375,13 +390,17 @@ local_annex_size = simpleStat "local annex size" $
 
 -- "remote" is in the name for JSON backwards-compatibility
 repo_annex_keys :: UUID -> Stat
-repo_annex_keys u = stat "remote annex keys" $ json show $
-	countKeys <$> cachedRemoteData u
+repo_annex_keys u = stat "remote annex keys" $ \d ->
+	cachedRemoteData u >>= \case
+		Right rd -> json show (pure (countKeys rd)) d
+		Left n-> json id (pure n) d
 
 -- "remote" is in the name for JSON backwards-compatibility
 repo_annex_size :: UUID -> Stat
 repo_annex_size u = simpleStat "remote annex size" $
-	showSizeKeys =<< cachedRemoteData u
+	cachedRemoteData u >>= \case
+		Right d -> showSizeKeys d
+		Left n -> pure n
 
 known_annex_files :: Bool -> Stat
 known_annex_files isworktree = 
@@ -404,7 +423,7 @@ bad_data_size :: Stat
 bad_data_size = staleSize "bad keys size" gitAnnexBadDir
 
 key_size :: Key -> Stat
-key_size k = simpleStat "size" $ showSizeKeys $ foldKeys [k]
+key_size k = simpleStat "size" $ showSizeKeys $ addKey k emptyKeyInfo
 
 key_name :: Key -> Stat
 key_name k = simpleStat "key" $ pure $ serializeKey k
@@ -424,7 +443,7 @@ bloom_info = simpleStat "bloom filter size" $ do
 	-- Two bloom filters are used at the same time when running
 	-- git-annex unused, so double the size of one.
 	sizer <- mkSizer
-	size <- sizer memoryUnits False . (* 2) . fromIntegral . fst <$>
+	size <- sizer committeeUnits False . (* 2) . fromIntegral . fst <$>
 		lift bloomBitsHashes
 
 	return $ size ++ note
@@ -441,17 +460,17 @@ transfer_list = stat desc $ nojson $ lift $ do
   where
 	desc = "transfers in progress"
 	line uuidmap t i = unwords
-		[ formatDirection (transferDirection t) ++ "ing"
-		, actionItemDesc $ mkActionItem
+		[ fromRawFilePath (formatDirection (transferDirection t)) ++ "ing"
+		, fromRawFilePath $ actionItemDesc $ mkActionItem
 			(transferKey t, associatedFile i)
 		, if transferDirection t == Upload then "to" else "from"
 		, maybe (fromUUID $ transferUUID t) Remote.name $
 			M.lookup (transferUUID t) uuidmap
 		]
-	jsonify t i = object $ map (\(k, v) -> (packString k, v)) $
+	jsonify t i = object $ map (\(k, v) -> (textKey (packString k), v)) $
 		[ ("transfer", toJSON' (formatDirection (transferDirection t)))
 		, ("key", toJSON' (transferKey t))
-		, ("file", toJSON' afile)
+		, ("file", toJSON' (fromRawFilePath <$> afile))
 		, ("remote", toJSON' (fromUUID (transferUUID t) :: String))
 		]
 	  where
@@ -461,7 +480,7 @@ disk_size :: Stat
 disk_size = simpleStat "available local disk space" $
 	calcfree
 		<$> (lift $ annexDiskReserve <$> Annex.getGitConfig)
-		<*> (lift $ inRepo $ getDiskFree . gitAnnexDir)
+		<*> (lift $ inRepo $ getDiskFree . fromRawFilePath . gitAnnexDir)
 		<*> mkSizer
   where
 	calcfree reserve (Just have) sizer = unwords
@@ -502,9 +521,9 @@ reposizes_stats = stat desc $ nojson $ do
 		. M.toList
 		<$> cachedRepoData
 	let maxlen = maximum (map (length . snd) l)
-	descm <- lift uuidDescriptions
+	descm <- lift Remote.uuidDescriptions
 	-- This also handles json display.
-	s <- lift $ prettyPrintUUIDsWith (Just "size") desc descm (Just . show) $
+	s <- lift $ Remote.prettyPrintUUIDsWith (Just "size") desc descm (Just . show) $
 		map (\(u, sz) -> (u, Just $ mkdisp sz maxlen)) l
 	return $ countRepoList (length l) s
   where
@@ -519,38 +538,52 @@ reposizes_total :: Stat
 reposizes_total = simpleStat "combined size of repositories containing these files" $
 	showSizeKeys . mconcat . M.elems =<< cachedRepoData
 
-cachedPresentData :: StatState KeyData
+cachedPresentData :: StatState KeyInfo
 cachedPresentData = do
 	s <- get
 	case presentData s of
 		Just v -> return v
 		Nothing -> do
-			v <- foldKeys <$> lift (getKeysPresent InRepository)
+			matcher <- lift getKeyOnlyMatcher
+			v <- foldl' (flip addKey) emptyKeyInfo
+				<$> lift (listKeys' InAnnex (matchOnKey matcher))
 			put s { presentData = Just v }
 			return v
 
-cachedRemoteData :: UUID -> StatState KeyData
+cachedRemoteData :: UUID -> StatState (Either String KeyInfo)
 cachedRemoteData u = do
 	s <- get
 	case M.lookup u (repoData s) of
-		Just v -> return v
+		Just v -> return (Right v)
 		Nothing -> do
+			matcher <- lift getKeyOnlyMatcher
 			let combinedata d uk = finishCheck uk >>= \case
 				Nothing -> return d
-				Just k -> return $ addKey k d
-			v <- lift $ foldM combinedata emptyKeyData
-				=<< loggedKeysFor' u
-			put s { repoData = M.insert u v (repoData s) }
-			return v
+				Just k -> ifM (matchOnKey matcher k)
+					( return (addKey k d)
+					, return d
+					)
+			lift (loggedKeysFor' u) >>= \case
+				Just (ks, cleanup) -> do
+					v <- lift $ foldM combinedata emptyKeyInfo ks
+					liftIO $ void cleanup
+					put s { repoData = M.insert u v (repoData s) }
+					return (Right v)
+				Nothing -> return (Left "not available in this read-only repository with unmerged git-annex branches")
 
-cachedReferencedData :: StatState KeyData
+cachedReferencedData :: StatState KeyInfo
 cachedReferencedData = do
 	s <- get
 	case referencedData s of
 		Just v -> return v
 		Nothing -> do
+			matcher <- lift getKeyOnlyMatcher
+			let combinedata k _f d = ifM (matchOnKey matcher k)
+				( return (addKey k d)
+				, return d
+				)
 			!v <- lift $ Command.Unused.withKeysReferenced
-				emptyKeyData addKey
+				emptyKeyInfo combinedata
 			put s { referencedData = Just v }
 			return v
 
@@ -559,7 +592,7 @@ cachedNumCopiesStats :: StatState (Maybe NumCopiesStats)
 cachedNumCopiesStats = numCopiesStats <$> get
 
 -- currently only available for directory info
-cachedRepoData :: StatState (M.Map UUID KeyData)
+cachedRepoData :: StatState (M.Map UUID KeyInfo)
 cachedRepoData = repoData <$> get
 
 getDirStatInfo :: InfoOptions -> FilePath -> Annex StatInfo
@@ -571,9 +604,9 @@ getDirStatInfo o dir = do
 			(update matcher fast)
 	return $ StatInfo (Just presentdata) (Just referenceddata) repodata (Just numcopiesstats) o
   where
-	initial = (emptyKeyData, emptyKeyData, emptyNumCopiesStats, M.empty)
+	initial = (emptyKeyInfo, emptyKeyInfo, emptyNumCopiesStats, M.empty)
 	update matcher fast key file vs@(presentdata, referenceddata, numcopiesstats, repodata) =
-		ifM (matcher $ MatchingFile $ FileInfo file file)
+		ifM (matcher $ MatchingFile $ FileInfo file file (Just key))
 			( do
 				!presentdata' <- ifM (inAnnex key)
 					( return $ addKey key presentdata
@@ -593,62 +626,69 @@ getDirStatInfo o dir = do
 getTreeStatInfo :: InfoOptions -> Git.Ref -> Annex (Maybe StatInfo)
 getTreeStatInfo o r = do
 	fast <- Annex.getState Annex.fast
-	(ls, cleanup) <- inRepo $ LsTree.lsTree LsTree.LsTreeRecursive r
-	(presentdata, referenceddata, repodata) <- go fast ls initial
+	-- git lstree filenames start with a leading "./" that prevents
+	-- matching, and also things like --include are supposed to
+	-- match relative to the current directory, which does not make
+	-- sense when matching against files in some arbitrary tree.
+	matcher <- getKeyOnlyMatcher
+	(ls, cleanup) <- inRepo $ LsTree.lsTree
+		LsTree.LsTreeRecursive
+		(LsTree.LsTreeLong False)
+		r
+	(presentdata, referenceddata, repodata) <- go fast matcher ls initial
 	ifM (liftIO cleanup)
 		( return $ Just $
 			StatInfo (Just presentdata) (Just referenceddata) repodata Nothing o
 		, return Nothing
 		)
   where
-	initial = (emptyKeyData, emptyKeyData, M.empty)
-	go _ [] vs = return vs
-	go fast (l:ls) vs@(presentdata, referenceddata, repodata) = do
-		mk <- catKey (LsTree.sha l)
-		case mk of
-			Nothing -> go fast ls vs
-			Just key -> do
-				!presentdata' <- ifM (inAnnex key)
-					( return $ addKey key presentdata
-					, return presentdata
-					)
-				let !referenceddata' = addKey key referenceddata
-				!repodata' <- if fast
-					then return repodata
-					else do
-						locs <- Remote.keyLocations key
-						return (updateRepoData key locs repodata)
-				go fast ls $! (presentdata', referenceddata', repodata')
+	initial = (emptyKeyInfo, emptyKeyInfo, M.empty)
+	go _ _ [] vs = return vs
+	go fast matcher (l:ls) vs@(presentdata, referenceddata, repodata) =
+		catKey (LsTree.sha l) >>= \case
+			Nothing -> go fast matcher ls vs
+			Just key -> ifM (matchOnKey matcher key)
+				( do
+					!presentdata' <- ifM (inAnnex key)
+						( return $ addKey key presentdata
+						, return presentdata
+						)
+					let !referenceddata' = addKey key referenceddata
+					!repodata' <- if fast
+						then return repodata
+						else do
+							locs <- Remote.keyLocations key
+							return (updateRepoData key locs repodata)
+					go fast matcher ls $! (presentdata', referenceddata', repodata')
+				, go fast matcher ls vs
+				)
 
-emptyKeyData :: KeyData
-emptyKeyData = KeyData 0 0 0 M.empty
+emptyKeyInfo :: KeyInfo
+emptyKeyInfo = KeyInfo 0 0 0 M.empty
 
 emptyNumCopiesStats :: NumCopiesStats
 emptyNumCopiesStats = NumCopiesStats M.empty
 
-foldKeys :: [Key] -> KeyData
-foldKeys = foldl' (flip addKey) emptyKeyData
-
-addKey :: Key -> KeyData -> KeyData
-addKey key (KeyData count size unknownsize backends) =
-	KeyData count' size' unknownsize' backends'
+addKey :: Key -> KeyInfo -> KeyInfo
+addKey key (KeyInfo count size unknownsize backends) =
+	KeyInfo count' size' unknownsize' backends'
   where
 	{- All calculations strict to avoid thunks when repeatedly
 	 - applied to many keys. -}
 	!count' = count + 1
-	!backends' = M.insertWith (+) (keyVariety key) 1 backends
+	!backends' = M.insertWith (+) (fromKey keyVariety key) 1 backends
 	!size' = maybe size (+ size) ks
 	!unknownsize' = maybe (unknownsize + 1) (const unknownsize) ks
-	ks = keySize key
+	ks = fromKey keySize key
 
-updateRepoData :: Key -> [UUID] -> M.Map UUID KeyData -> M.Map UUID KeyData
+updateRepoData :: Key -> [UUID] -> M.Map UUID KeyInfo -> M.Map UUID KeyInfo
 updateRepoData key locs m = m'
   where
 	!m' = M.unionWith (\_old new -> new) m $
 		M.fromList $ zip locs (map update locs)
-	update loc = addKey key (fromMaybe emptyKeyData $ M.lookup loc m)
+	update loc = addKey key (fromMaybe emptyKeyInfo $ M.lookup loc m)
 
-updateNumCopiesStats :: FilePath -> NumCopiesStats -> [UUID] -> Annex NumCopiesStats
+updateNumCopiesStats :: RawFilePath -> NumCopiesStats -> [UUID] -> Annex NumCopiesStats
 updateNumCopiesStats file (NumCopiesStats m) locs = do
 	have <- trustExclude UnTrusted locs
 	!variance <- Variance <$> numCopiesCheck' file (-) have
@@ -656,7 +696,7 @@ updateNumCopiesStats file (NumCopiesStats m) locs = do
 	let !ret = NumCopiesStats m'
 	return ret
 
-showSizeKeys :: KeyData -> StatState String
+showSizeKeys :: KeyInfo -> StatState String
 showSizeKeys d = do
 	sizer <- mkSizer
 	return $ total sizer ++ missingnote
@@ -668,7 +708,7 @@ showSizeKeys d = do
 			"+ " ++ show (unknownSizeKeys d) ++
 			" unknown size"
 
-staleSize :: String -> (Git.Repo -> FilePath) -> Stat
+staleSize :: String -> (Git.Repo -> RawFilePath) -> Stat
 staleSize label dirspec = go =<< lift (dirKeys dirspec)
   where
 	go [] = nostat
@@ -680,8 +720,8 @@ staleSize label dirspec = go =<< lift (dirKeys dirspec)
 			return $ sizer storageUnits False size
 	keysizes keys = do
 		dir <- lift $ fromRepo dirspec
-		liftIO $ forM keys $ \k -> catchDefaultIO 0 $
-			getFileSize (dir </> keyFile k)
+		liftIO $ forM keys $ \k -> 
+			catchDefaultIO 0 $ getFileSize (dir P.</> keyFile k)
 
 aside :: String -> String
 aside s = " (" ++ s ++ ")"
@@ -694,3 +734,20 @@ mkSizer = ifM (bytesOption . infoOptions <$> get)
 	( return (const $ const show)
 	, return roughSize
 	)
+			
+getKeyOnlyMatcher :: Annex (MatchInfo -> Annex Bool)
+getKeyOnlyMatcher = do
+	whenM (Limit.introspect matchNeedsFileName) $ do
+		warning "File matching options cannot be applied when getting this info."
+		giveup "Unable to continue."
+	Limit.getMatcher
+
+matchOnKey :: (MatchInfo -> Annex Bool) -> Key -> Annex Bool
+matchOnKey matcher k = matcher $ MatchingInfo $ ProvidedInfo
+	{ providedFilePath = Nothing
+	, providedKey = Just k
+	, providedFileSize = Nothing
+	, providedMimeType = Nothing
+	, providedMimeEncoding = Nothing
+	, providedLinkType = Nothing
+	}
